import asyncio
import aiohttp
from typing import List, Dict, Any, Optional
from openai_manager.auth_manager import OpenAIAuthManager, OpenAIAuth, APIRequest, REQUESTS_PER_MIN_LIMIT, TOKENS_PER_MIN_LIMIT, GLOBAL_NUM_REQUEST_LIMIT
import os
import time
from openai_manager.exceptions import NoAvailableAuthException, AttemptsExhaustedException
from openai_manager.utils import num_tokens_consumed_from_request, logger, str2bool
from tqdm import tqdm
from openai.util import convert_to_openai_object

# notice loading custom YAML config will overwrite these envvars

COROTINE_PER_AUTH = int(os.getenv("COROTINE_PER_AUTH", 3))
ATTEMPTS_PER_PROMPT = int(os.getenv("ATTEMPTS_PER_PROMPT", 5))
RATELIMIT_AFTER_SUBMISSION = str2bool(
    os.getenv("RATELIMIT_AFTER_SUBMISSION", "True"))

if COROTINE_PER_AUTH == 1:  # single thread, no need for `after_submission_update`
    RATELIMIT_AFTER_SUBMISSION = False


async def consume_submit(q: asyncio.Queue, auth: OpenAIAuth, results: List[dict],
                         auth_manager, thread_id: int, session, api_endpoint='completions', pbar: Optional[tqdm] = None):
    while True:
        logger.debug(
            f"thread {thread_id} auth {auth.auth_index} waiting for submission...")
        item = await q.get()
        item: APIRequest
        logger.debug(
            f"thread {thread_id} auth {auth.auth_index} received submission {item}")
        # check if exhausted attempts
        if item.attempts_left == 0:
            raise AttemptsExhaustedException(
                f"Request {item} exhausted all attempts.")

        item.attempts_left -= 1

        # --- below are our second-level rate limit controller, should be more strict! ---
        before_submission_time = time.time()
        if auth.ratelimit_tracker.next_request_time != 0:
            auth.ratelimit_tracker.next_request_time = max(
                auth.ratelimit_tracker.next_request_time +
                auth.ratelimit_tracker.seconds_per_request,
                auth.ratelimit_tracker.next_request_time +
                auth.ratelimit_tracker.seconds_per_token * item.token_consumption
            )
        else:
            auth.ratelimit_tracker.next_request_time = before_submission_time  # first request
        logger.info(
            f"thread {thread_id} auth {auth.auth_index} update next_request_time to {auth.ratelimit_tracker.next_request_time}")
        if before_submission_time < auth.ratelimit_tracker.next_request_time:
            logger.info(
                f"thread {thread_id} auth {auth.auth_index} waiting for next request time..., before_submission_time: {before_submission_time}, next_request_time: {auth.ratelimit_tracker.next_request_time}")
            await asyncio.sleep(auth.ratelimit_tracker.next_request_time - before_submission_time)

        response = await item.call_API_pure(
            request_url=auth_manager.endpoints[api_endpoint],
            request_header={
                'Authorization': f'Bearer {auth.api_key}'},
            retry_queue=q,
            auth=auth,
            # session=auth_manager.session,
            session=session,
            thread_id=thread_id,
        )
        # we update next_request_time when receiving the response, as sometimes response takes very long
        if RATELIMIT_AFTER_SUBMISSION:
            after_submission_time = time.time()
            auth.ratelimit_tracker.next_request_time = max(
                after_submission_time +
                auth.ratelimit_tracker.seconds_per_request,
                after_submission_time +
                auth.ratelimit_tracker.seconds_per_token * item.token_consumption
            )
            logger.info(
                f"after submission, thread {thread_id} auth {auth.auth_index} update next_request_time to {auth.ratelimit_tracker.next_request_time}")
            logger.info(
                f"thread {thread_id} auth {auth.auth_index} received request {item.task_id} at {time.time()}, next_request_time is {auth.ratelimit_tracker.next_request_time}")

        if pbar is not None:
            pbar.update()
            if response['error']:
                pbar.total += 1   # put error back to queue
                pbar.refresh()

        if response['error']:
            logger.warning(
                f"thread {thread_id} auth {auth.auth_index} experience error...")

        results.append(response)  # remember to sort by task_id when returning

        q.task_done()


async def batch_submission(auth_manager: OpenAIAuthManager, input: List[Any], pbar=None,
                           api_endpoint='completions',
                           **kwargs) -> List[Dict[str, Any]]:
    if len(auth_manager.auths) == 0:
        raise NoAvailableAuthException(f"No available OpenAI Auths!")
    ret = []
    q = asyncio.Queue()  # storing APIRequest
    for i in input:
        # all unused kwargs goes here!
        request_json = i
        # TODO: automatic batching for a large number of prompts
        token_consumption = num_tokens_consumed_from_request(
            request_json=request_json,
            api_endpoint=api_endpoint,
            token_encoding_name='cl100k_base',
        )
        await q.put(APIRequest(
            task_id=next(auth_manager.task_id_generator),
            request_json=request_json,
            token_consumption=token_consumption,
            attempts_left=ATTEMPTS_PER_PROMPT,
        ))

    consumers = []
    thread_id = 0
    # move connector and session here to avoid 'Timeout context manager should be used inside a task'
    async with aiohttp.TCPConnector(limit=GLOBAL_NUM_REQUEST_LIMIT) as connector:
        async with aiohttp.ClientSession(connector=connector) as session:
            for auth in auth_manager.auths:
                consumers.extend([asyncio.create_task(consume_submit(
                    q, auth, ret, auth_manager, thread_id + j, session, api_endpoint=api_endpoint, pbar=pbar))
                    for j in range(COROTINE_PER_AUTH)])
                thread_id += COROTINE_PER_AUTH
            await q.join()
    # for auth in auth_manager.auths:
    #     consumers.extend([asyncio.create_task(consume_submit(
    #         q, auth, ret, auth_manager, thread_id + j, api_endpoint=api_endpoint, pbar=pbar))
    #         for j in range(COROTINE_PER_AUTH)])
    #     thread_id += COROTINE_PER_AUTH
    await q.join()  # wait until queue is empty
    for c in consumers:  # shutdown all consumers
        c.cancel()
    # gather results, ignore errors
    return list(sorted([result for result in ret if not result['error']], key=lambda x: x['task_id']))


def sync_batch_submission(auth_manager, prompt: Optional[List[str]] = None, input: Optional[List[str]] = None,
                          messages: Optional[List[dict]] = None, debug=False, no_tqdm=False, return_openai_response=False, **kwargs):
    loop = asyncio.get_event_loop()
    if prompt is not None:  # completion request
        input_lst = [{"prompt": p, **kwargs} for p in prompt]
        api_endpoint = 'completions'
    elif input is not None:  # embedding request
        input_lst = [{"input": i, **kwargs} for i in input]
        api_endpoint = 'embeddings'
    elif messages is not None:  # chatcompletion request
        input_lst = [{"messages": m, **kwargs} for m in messages]
        api_endpoint = 'chat_completions'
    else:
        raise ValueError("At lease one type of input should be provided!")
    pbar = tqdm(total=len(input_lst)) if not no_tqdm else None
    responses_with_error_and_task_id = loop.run_until_complete(
        batch_submission(auth_manager, input_lst, pbar=pbar, api_endpoint=api_endpoint))
    if not pbar:
        pbar.close()
    # keep 'response' only
    # TODO: if gets automatic batching, remember to restore them back
    if not return_openai_response:
        return [response['response'] for response in responses_with_error_and_task_id]
    else:
        return [convert_to_openai_object(response['response']) for response in responses_with_error_and_task_id]


if __name__ == '__main__':
    auth_manager = OpenAIAuthManager()
    print(sync_batch_submission(auth_manager, [
          "Hello world!"] * 30, max_tokens=20, model='text-ada-001'))



import unittest
import openai
# import multiprocessing
# import os

# for multithread uvicorn solution
import uvicorn
import contextlib
import time
import threading


class Server(uvicorn.Server):
    def install_signal_handlers(self):
        pass

    @contextlib.contextmanager
    def run_in_thread(self):
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            while not self.started:
                time.sleep(1e-3)
            yield
        finally:
            self.should_exit = True
            thread.join()

# def create_serving_process():
#     """A ugly os patch to run command in a separate process"""
#     def _run():
#         os.system("python -m openai_manager.serving --port 8000 --host localhost --api_key sk-test_openai_manager_serving")
#     p = multiprocessing.Process(target=_run)
#     p.start()
#     return p


class TestServing(unittest.TestCase):
    def setUp(self) -> None:
        # self.serving_process = create_serving_process()
        from openai_manager.serving import app
        openai.api_base = "http://localhost:8000/v1"
        openai.api_key = "sk-test_openai_manager_serving"
        app.api_key = "sk-test_openai_manager_serving"
        config = uvicorn.Config(
            app=app, host="localhost", port=8000, log_level="info")
        self.server = Server(config=config)

    def test_openai_create(self):
        with self.server.run_in_thread():
            prompt = ["Once upon a time, "] * 5
            response = openai.Completion.create(
                model="text-ada-001",
                prompt=prompt,
                max_tokens=10,
            )
            print(response)
            self.assertEqual(len(response['choices']), 5)


if __name__ == '__main__':
    unittest.main()



# light-weight API wrapper for identical call signatures with official openai-python
from openai_manager.producer_consumer import sync_batch_submission
from openai_manager import GLOBAL_MANAGER


class Completion:
    @classmethod
    def create(cls, **params):
        prompt = params.pop('prompt')  # ensure a required field
        if not isinstance(prompt, list):
            prompt = [prompt]
        return sync_batch_submission(GLOBAL_MANAGER, prompt=prompt, **params)


class Embedding:
    @classmethod
    def create(cls, **params):
        input = params.pop('input')
        if not isinstance(input, list):
            input = [input]
        return sync_batch_submission(GLOBAL_MANAGER, input=input, **params)


class ChatCompletion:
    @classmethod
    def create(cls, **params):
        messages = params.pop('messages', None)
        if not isinstance(messages, list):
            raise TypeError(f"messages must be a list of dict, got {type(messages)}")
        # when provided in openai style, wrap it with a list, despite this will not allow parallellization
        if isinstance(messages[0], dict):
            messages = [messages]
        return sync_batch_submission(GLOBAL_MANAGER, messages=messages, **params)



class ManagerException(Exception):
    def __init__(self, message) -> None:
        super().__init__(message)


class NoAvailableAuthException(ManagerException):
    pass


class AttemptsExhaustedException(ManagerException):
    pass



# pylint: skip-file
# this script inits a FastAPI endpoint and an OpenAIManager
# allowing distributing openai requests to multiple keys from all official openai api implementations
import uvicorn
import argparse
from fastapi import FastAPI, Request
from openai_manager import GLOBAL_MANAGER
from openai_manager.producer_consumer import batch_submission

app = FastAPI()


@app.post("/v1/completions")
async def completions(request: Request):
    # validate first
    key = request.headers.get('authorization', None)
    print(f"received request: {request}")
    if key != f'Bearer {app.api_key}':  # covered None case
        return {
            "error": {
                "message": f"Incorrect API key provided: {key}. You should set the key in openai-manager via command.",
                "type": "invalid_request_error",
                "param": None,
                "code": "invalid_api_key"
            }
        }
    request_json = await request.json()
    prompt = request_json.pop('prompt')
    if not isinstance(prompt, list):
        prompt = [prompt]

    input_lst = [{"prompt": p, **request_json} for p in prompt]
    responses_with_error_and_task_id = await batch_submission(GLOBAL_MANAGER, input_lst, pbar=None, api_endpoint='completions')
    # keep response only, and format into openai api response
    response = [response['response']
                for response in responses_with_error_and_task_id]
    assert len(response) > 0, "Empty response"
    # we keep the first request's `id`, `object`, `created`, `model`; sum over all the usage, and combine the `choices`
    return {
        'id': response[0]['id'],
        'object': response[0]['object'],
        'created': response[0]['created'],
        'model': response[0]['model'],
        # `index` kept; our implementation ensures ascending order
        'choices': [c for res in response for c in res['choices']],
        'usage': {
            'prompt_tokens': sum([res['usage']['prompt_tokens'] for res in response]),
            'completion_tokens': sum([res['usage']['completion_tokens'] for res in response]),
            'total_tokens': sum([res['usage']['total_tokens'] for res in response]),
        }
    }


@app.post("/v1/completions/embeddings")
async def embeddings(request: Request):
    pass


@app.post("/v1/chat/completions")
async def chat_completions(request: Request):
    pass


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, default=8000)
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--reload", action='store_true', default=False)
    parser.add_argument("--custom_config", type=str, default=None)
    # this api key is used to do self validation, not for openai requests
    parser.add_argument("--api_key", type=str, default="sk-123")
    # post args checking
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()
    GLOBAL_MANAGER.append_auth_from_config(config_path=args.custom_config)
    app.api_key = args.api_key
    uvicorn.run(app, host=args.host, port=args.port, reload=args.reload)



# pylint: skip-file
# avoid parially initialized module error
from openai_manager.auth_manager import OpenAIAuthManager  # noqa
GLOBAL_MANAGER = OpenAIAuthManager()  # noqa


from openai_manager.api_resources import Completion, Embedding, ChatCompletion  # noqa
import logging
import os


logging.basicConfig(level=int(os.getenv("OPENAI_LOG_LEVEL", logging.WARNING)))
logger = logging.getLogger(__name__)
logger.debug(f"Logger level: {logger.level}")


def append_auth_from_config(config_path=None, config_dicts=None):
    GLOBAL_MANAGER.append_auth_from_config(config_path, config_dicts)


__all__ = [
    "Completion", 
    "Embedding", 
    "ChatCompletion"
]

name = "openai-manager"



import tiktoken
import warnings
import functools
import os
import logging
import argparse
import time
from typing import List, Dict

logging.basicConfig(level=int(os.getenv("OPENAI_LOG_LEVEL", logging.WARNING)))
logger = logging.getLogger(__name__)
logger.debug(f"Logger level: {logger.level}")


def timeit(func):
    @functools.wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(
            f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')
        return result
    return timeit_wrapper


def str2bool(s):
    if s.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif s.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)
    return new_func


def num_tokens_consumed_from_request(
    request_json: dict,
    api_endpoint: str,
    token_encoding_name: str,
):
    """Count the number of tokens in the request. Only supports completion and embedding requests."""
    encoding = tiktoken.get_encoding(token_encoding_name)
    # if completions request, tokens = prompt + n * max_tokens
    if api_endpoint == "completions":
        prompt = request_json["prompt"]
        max_tokens = request_json.get("max_tokens", 15)
        n = request_json.get("n", 1)
        completion_tokens = n * max_tokens
        if isinstance(prompt, str):  # single prompt
            prompt_tokens = len(encoding.encode(prompt))
            num_tokens = prompt_tokens + completion_tokens
            return num_tokens
        elif isinstance(prompt, list):  # multiple prompts
            prompt_tokens = sum([len(encoding.encode(p)) for p in prompt])
            num_tokens = prompt_tokens + completion_tokens * len(prompt)
            return num_tokens
        else:
            raise TypeError(
                'Expecting either string or list of strings for "prompt" field in completion request')
    # if embeddings request, tokens = input tokens
    elif api_endpoint == "embeddings":
        input = request_json["input"]
        if isinstance(input, str):  # single input
            num_tokens = len(encoding.encode(input))
            return num_tokens
        elif isinstance(input, list):  # multiple inputs
            num_tokens = sum([len(encoding.encode(i)) for i in input])
            return num_tokens
        else:
            raise TypeError(
                'Expecting either string or list of strings for "inputs" field in embedding request')
    # more logic needed to support other API calls (e.g., edits, inserts, DALL-E)
    # if chat_completions request, tokens = all content input tokens
    elif api_endpoint == 'chat_completions':
        num_tokens = 0
        max_tokens = request_json.get("max_tokens", 15)
        n = request_json.get("n", 1)
        completion_tokens = n * max_tokens
        # note: future models may deviate from this
        if request_json['model'].startswith("gpt-3.5-turbo"):
            # chat completions are also in list
            for message in request_json['messages']:
                # every message follows <im_start>{role/name}\n{content}<im_end>\n
                num_tokens += 4
                for key, value in message.items():
                    num_tokens += len(encoding.encode(value))
                    if key == "name":  # if there's a name, the role is omitted
                        num_tokens += -1  # role is always required and always 1 token
            num_tokens += 2  # every reply is primed with <im_start>assistant
        return num_tokens + completion_tokens
    else:
        raise NotImplementedError(
            f'API endpoint "{api_endpoint}" not implemented in this script')



import unittest
import openai_manager

class TestChatCompletion(unittest.TestCase):
    def test_completion(self):
        response = openai_manager.Completion.create(
            model="text-ada-001",
            prompt=["Once upon a time, "] * 10,
            max_tokens=20,
        )
        print(response)
        self.assertEqual(len(response), 10)
        self.assertTrue(all(isinstance(res, dict) for res in response))
    
    def test_chat_completion(self):
        response = openai_manager.ChatCompletion.create(model='gpt-3.5-turbo',
                                                        messages=[
                                                            [{"role": "user", "content": "Hello!"}],
                                                            [{"role": "user", "content": "Hello!"}, {"role": "assistant", "content": "Hello there!"}, {
                                                                "role": "user", "content": "Who are you?"}]
                                                        ])
        print(response)
        # typical response would be: 
        # [{'id': 'chatcmpl-7035RQGB86hJSuqc0yICkegAMXIi0', 'object': 'chat.completion', 'created': 1680246221, 'model': 'gpt-3.5-turbo-0301', 'usage': {'prompt_tokens': 10, 'completion_tokens': 9, 'total_tokens': 19}, 'choices': [{'message': {'role': 'assistant', 'content': 'Hello! How can I assist you today?'}, 'finish_reason': 'stop', 'index': 0}]}, {'id': 'chatcmpl-7035Wyy6OSU3nzsDPhM5cWXl2mtgD', 'object': 'chat.completion', 'created': 1680246226, 'model': 'gpt-3.5-turbo-0301', 'usage': {'prompt_tokens': 27, 'completion_tokens': 31, 'total_tokens': 58}, 'choices': [{'message': {'role': 'assistant', 'content': 'I am an AI language model designed by OpenAI. You can ask me questions or ask for assistance with various tasks. How may I assist you today?'}, 'finish_reason': 'stop', 'index': 0}]}]
        self.assertEqual(len(response), 2)
        self.assertTrue(all(isinstance(res, dict) for res in response))
        
if __name__ == '__main__':
    unittest.main()
    


from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any
import asyncio
import aiohttp
import time
import os
import traceback
from openai_manager.utils import logger


# notice loading custom YAML config will overwrite these envvars
GLOBAL_NUM_REQUEST_LIMIT = int(os.getenv(
    "OPENAI_GLOBAL_NUM_REQUEST_LIMIT", 500))
PROMPTS_PER_ASYNC_BATCH = int(os.getenv(
    "OPENAI_PROMPTS_PER_ASYNC_BATCH", 1000))
# 20 requests per minute in `code-davinci-002`, we set it as default
REQUESTS_PER_MIN_LIMIT = int(os.getenv(
    "OPENAI_REQUESTS_PER_MIN_LIMIT", 10))
# 40,000 tokens per minute in `code-davinci-002`, we set it as default
TOKENS_PER_MIN_LIMIT = int(os.getenv("TOKENS_PER_MIN_LIMIT", 40_000))


@dataclass
class StatusTracker:
    # from: https://github.com/openai/openai-cookbook/blob/2f5e350bbe66a418184899b0e12f182dbb46a156/examples/api_request_parallel_processor.py
    """
    Stores metadata about each auth's progress.
    Will trigger a cool-off period if rate limits are hit by asyncio.sleep()
    """
    num_tasks_started: int = 0
    num_tasks_in_progress: int = 0
    num_tasks_succeeded: int = 0
    num_tasks_failed: int = 0
    num_rate_limit_errors: int = 0
    num_api_errors: int = 0  # excluding rate limit errors, counted above
    num_other_errors: int = 0
    # used to cool off after hitting rate limits
    time_of_last_rate_limit_error: int = 0


@dataclass
class RateLimitTracker:
    """
    Keep track of each auth's rate limit status.
    OpenAI applys rate limit in both `requests per minute` and `tokens per minute`.
    RateLimit Tracker should use **second-level** tracker for requests in practice, 
    as OpenAI told in https://platform.openai.com/docs/guides/rate-limits/overview
    """
    next_request_time: int = 0
    seconds_to_pause_after_rate_limit_error: int = 10
    seconds_to_sleep_each_loop: int = 0.001
    available_request_capacity: int = REQUESTS_PER_MIN_LIMIT
    available_token_capacity: int = TOKENS_PER_MIN_LIMIT
    # epsilon_time: float = 0.001  # 1ms delay for corotine rotation
    last_update_time: int = time.time()

    def __post_init__(self):
        # a request at T, next request should be at T + (60 / available_request_capacity)
        self.seconds_per_request = 60 / self.available_request_capacity
        # X tokens at T, next request should be at T + (60 / available_token_capacity) * X
        self.seconds_per_token = 60 / self.available_token_capacity


@dataclass
class OpenAIAuth:
    auth_index: int
    api_key: str
    status_tracker: StatusTracker = field(default_factory=StatusTracker)
    ratelimit_tracker: RateLimitTracker = field(
        default_factory=RateLimitTracker)
    proxy: Optional[str] = None
    is_okay: bool = True
    in_use: bool = False


def task_id_generator_function():
    i = 0
    while True:
        yield i
        i += 1


async def create_session():
    return aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(limit=GLOBAL_NUM_REQUEST_LIMIT)
    )


class OpenAIAuthManager:
    # when importing openai_manager, this class will be initialized
    endpoints = {
        'completions': 'https://api.openai.com/v1/completions',
        'embeddings': 'https://api.openai.com/v1/embeddings',
        'chat_completions': 'https://api.openai.com/v1/chat/completions'
    }
    task_id_generator = task_id_generator_function()

    def __init__(self) -> None:
        # ".env" overrides envvars
        env_source = self.parse_dict_for_env_file() if os.path.exists('.env') else os.environ
        self.auths = self.build_auth_from_envvars(dict_source=env_source)
        logger.warning(f"Loaded {len(self.auths)} OpenAI auths...")
        # self.session = asyncio.get_event_loop().run_until_complete(create_session())
        self.session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(limit=GLOBAL_NUM_REQUEST_LIMIT)
        )
        # DeprecationWarning: session was shared among corotinues, allowing keeping connection alive

    def parse_dict_for_env_file(self) -> Dict[str, str]:
        env_dict = dict()
        if os.path.exists('.env'):
            with open('.env', 'r') as f:
                env_lines = f.readlines()
                env_lines = [line.split('=') for line in env_lines]
                for env_line in env_lines:
                    if len(env_line) != 2:
                        # skip invalid lines
                        continue
                    env_dict[env_line[0].strip()] = env_line[1].strip()
        return env_dict

    def build_auth_from_envvars(self, dict_source) -> List[OpenAIAuth]:
        # read from config file and build auths
        # auth priority: .env > env var > config file;
        # return a list of auths
        auths = []
        auth_i = 0
        default_proxy = dict_source.get('OPENAI_API_PROXY', None)
        if not default_proxy:
            default_proxy = None
        else:
            logger.warning(f'default proxy is set to {default_proxy}')
        for env_key, env_value in dict_source.items():
            if env_key.startswith('OPENAI_API_KEY'):
                env_key = env_key[len('OPENAI_API_KEY'):]
                auths.append(OpenAIAuth(auth_index=auth_i,
                                        api_key=env_value,
                                        proxy=dict_source.get(
                                            f'OPENAI_API_PROXY{env_key}', default_proxy)  # global proxy overwrites all if not provided proxy
                                        ))
                auth_i += 1
        return auths

    def append_auth_from_config(self, config_path: Optional[str] = None, config_dicts: Optional[List[Dict[str, Any]]] = None):
        if config_path is not None:
            try:
                import yaml
            except ImportError:
                logger.warning(
                    f"pyyaml is not installed, run `pip install pyyaml` in your current environment.")
                return
            with open(config_path, 'r') as f:
                config = yaml.safe_load(f)
            for key, value in config.items():
                if key.startswith('auth'):
                    requests_per_min = value.pop(
                        'requests_per_min', REQUESTS_PER_MIN_LIMIT)
                    tokens_per_min = value.pop(
                        'tokens_per_min', TOKENS_PER_MIN_LIMIT)
                    self.auths.append(OpenAIAuth(auth_index=len(self.auths),
                                                 ratelimit_tracker=RateLimitTracker(
                        available_request_capacity=requests_per_min,
                        available_token_capacity=tokens_per_min),
                        **value
                    ))
        if config_dicts is not None:
            # madantory keys `api_key` / `proxy`, etc...
            self.auths.extend([OpenAIAuth(auth_index=i + len(self.auths), **config_dict)
                              for i, config_dict in enumerate(config_dicts)])

    def get_available_auth(self) -> OpenAIAuth:
        # return the first available auth
        for auth in self.auths:
            if auth.is_okay and not auth.in_use:
                auth.in_use = True
                return auth

    def __del__(self):
        # close the session to avoid aiohttp error
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.session.close())


@dataclass
class APIRequest:
    """Stores an API request's inputs, outputs, and other metadata. Contains a method to make an API call."""

    task_id: int
    request_json: dict
    token_consumption: int
    attempts_left: int
    result = []

    async def call_API_pure(
        self,
        request_url: str,
        request_header: dict,
        retry_queue: asyncio.Queue,
        auth: OpenAIAuth,
        session: aiohttp.ClientSession,
        thread_id: int,
    ):
        # `pure version` remove ratelimit checker and capacity updater
        logger.info(f"thread {thread_id} Starting request #{self.task_id}")
        error = None
        try:
            post_params = {
                'url': request_url,
                'headers': request_header,
                'json': self.request_json
            }
            if auth.proxy:
                post_params.update({'proxy': auth.proxy})

            async with session.post(**post_params) as response:
                # most request gets blocked here; so our restriction is not useful
                response = await response.json()
            if "error" in response:
                # TODO: 1. disable auth if auth is invalid
                # TODO: 2. put into retry_queue if it is a temporary error
                # TODO: 3. throw exceptions immediately if some permanent errors occur
                logger.warning(
                    f"Request {self.task_id} failed with error {response['error']}"
                )
                auth.status_tracker.num_api_errors += 1
                error = response
                if "Rate limit" in response["error"].get("message", ""):
                    logger.debug(
                        f"Set time_of_last_rate_limit_error to {time.time()}")
                    # each time when triggering rate limit, record capacity:
                    # logger.warning(
                    #     f"Rate limit triggered: capacity: {auth.ratelimit_tracker.available_request_capacity} (requests) / {auth.ratelimit_tracker.available_token_capacity} (tokens)")
                    logger.warning(
                        f'Rate limit triggered: thread {thread_id} auth {auth.auth_index} now {time.time()} next_request_time {auth.ratelimit_tracker.next_request_time}'
                    )
                    auth.status_tracker.time_of_last_rate_limit_error = time.time()
                    auth.status_tracker.num_rate_limit_errors += 1
                    # rate limit errors are counted separately
                    auth.status_tracker.num_api_errors -= 1

        except Exception as e:  # catching naked exceptions is bad practice, but in this case we'll log & save them
            logger.warning(
                f"Request {self.task_id} failed with Exception {e}")
            traceback.print_exc()
            auth.status_tracker.num_other_errors += 1
            error = e

        if error:
            self.result.append(error)
            retry_queue.put_nowait(self)
            if self.attempts_left == 0:
                logger.error(
                    f"Request {self.request_json} failed after all attempts. Saving errors: {self.result}")
                auth.status_tracker.num_tasks_in_progress -= 1
                auth.status_tracker.num_tasks_failed += 1
        else:
            auth.status_tracker.num_tasks_in_progress -= 1
            auth.status_tracker.num_tasks_succeeded += 1

        return {'response': response, 'error': error, 'task_id': self.task_id}



import setuptools

with open("README.md", "r") as fh:
    _LONG_DESCRIPTION = fh.read()


setuptools.setup(
    name="openai-manager",
    license='MIT',
    author="MrZilinXiao",
    version="1.0.0",
    author_email="me@mrxiao.net",
    description="Speed up your OpenAI requests by balancing prompts to multiple API keys.",
    long_description=_LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/MrZilinXiao/openai-manager",
    packages=setuptools.find_namespace_packages(),
    install_requires=["aiohttp", "openai", "tiktoken", "fastapi"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7'
)
